@@ -31,6 +31,11 @@
 extern int errno;
 #endif
 
+#ifndef BOOTSTRAP
+#include <selinux/selinux.h>
+#include <selinux/context.h>
+#endif
+
 #ifdef HAVE_UNISTD_H
 # include <unistd.h>
 #endif
@@ -133,6 +138,9 @@ struct input {
 
   const char *in_file_name;
 
+  /* Owner and mode to be set just before closing the file.  */
+  struct stat st;
+
   /* if NULL, none of the following are valid */
   FILE *fp;
 
@@ -549,7 +557,7 @@ static inline void
 flush_output(fp)
   FILE *fp;
 {
-  if (fp != stdout || unbuffered_output)
+  if (fp != stdout || unbuffered)
     ck_fflush(fp);
 }
 
@@ -708,9 +716,13 @@ open_next_file(name, input)
 
   if (in_place_extension)
     {
-      int input_fd, output_fd;
+      int input_fd;
       char *tmpdir, *p;
-      struct stat st;
+#ifndef BOOTSTRAP
+      security_context_t old_fscreatecon;
+      int reset_fscreatecon = 0;
+      memset (&old_fscreatecon, 0, sizeof (old_fscreatecon));
+#endif
 
       if (follow_symlinks)
 	input->in_file_name = follow_symlink (name);
@@ -720,7 +732,7 @@ open_next_file(name, input)
       /* get the base name */
       tmpdir = ck_strdup(input->in_file_name);
       if ((p = strrchr(tmpdir, '/')))
-	*(p + 1) = 0;
+	*p = 0;
       else
 	strcpy(tmpdir, ".");
 
@@ -728,28 +740,54 @@ open_next_file(name, input)
         panic(_("couldn't edit %s: is a terminal"), input->in_file_name);
 
       input_fd = fileno (input->fp);
-      fstat (input_fd, &st);
-      if (!S_ISREG (st.st_mode))
+      fstat (input_fd, &input->st);
+      if (!S_ISREG (input->st.st_mode))
         panic(_("couldn't edit %s: not a regular file"), input->in_file_name);
 
+#ifndef BOOTSTRAP
+      if (is_selinux_enabled ())
+	{
+          security_context_t con;
+	  if (getfilecon (input->in_file_name, &con) != -1)
+	    {
+	      /* Save and restore the old context for the sake of w and W
+		 commands.  */
+	      reset_fscreatecon = getfscreatecon (&old_fscreatecon) >= 0;
+	      if (setfscreatecon (con) < 0)
+		fprintf (stderr, _("%s: warning: failed to set default file creation context to %s: %s"),
+			 myname, con, strerror (errno));
+	      freecon (con);
+	    }
+	  else
+	    {
+	      if (errno != ENOSYS)
+		fprintf (stderr, _("%s: warning: failed to get security context of %s: %s"),
+			 myname, input->in_file_name, strerror (errno));
+	    }
+	}
+#endif
+
       output_file.fp = ck_mkstemp (&input->out_file_name, tmpdir, "sed");
       output_file.missing_newline = false;
       free (tmpdir);
 
+#ifndef BOOTSTRAP
+      if (reset_fscreatecon)
+	{
+	  setfscreatecon (old_fscreatecon);
+	  freecon (old_fscreatecon);
+	}
+#endif
+
       if (!output_file.fp)
         panic(_("couldn't open temporary file %s: %s"), input->out_file_name, strerror(errno));
-
-      output_fd = fileno (output_file.fp);
-#ifdef HAVE_FCHOWN
-      if (fchown (output_fd, st.st_uid, st.st_gid) == -1)
-        fchown (output_fd, -1, st.st_gid);
-#endif
-      copy_acl (input->in_file_name, input_fd,
-		input->out_file_name, output_fd,
-		st.st_mode);
     }
   else
-    output_file.fp = stdout;
+    {
+      if (input->fp && unbuffered)
+        setvbuf (input->fp, NULL, _IONBF, 0);
+      output_file.fp = stdout;
+    }
 }
 
 
@@ -766,9 +804,21 @@ closedown(input)
   if (in_place_extension && output_file.fp != NULL)
     {
       const char *target_name;
-      ck_fclose (output_file.fp);
+      int input_fd, output_fd;
 
       target_name = input->in_file_name;
+      input_fd = fileno (input->fp);
+      output_fd = fileno (output_file.fp);
+      copy_acl (input->in_file_name, input_fd,
+		input->out_file_name, output_fd,
+		input->st.st_mode);
+#ifdef HAVE_FCHOWN
+      if (fchown (output_fd, input->st.st_uid, input->st.st_gid) == -1)
+        fchown (output_fd, -1, input->st.st_gid);
+#endif
+
+      ck_fclose (input->fp);
+      ck_fclose (output_file.fp);
       if (strcmp(in_place_extension, "*") != 0)
         {
           char *backup_file_name = get_backup_file_name(target_name);
@@ -779,8 +829,9 @@ closedown(input)
       ck_rename (input->out_file_name, target_name, input->out_file_name);
       free (input->out_file_name);
     }
+  else
+    ck_fclose (input->fp);
 
-  ck_fclose (input->fp);
   input->fp = NULL;
 }
 
